<?php
App::uses('AppModel', 'Model');
App::uses('BlowfishPasswordHasher', 'Controller/Component/Auth');

/**
 * User Model
 *
 */
class User extends AppModel
{

    public $order = 'User.name ASC';
    public $actsAs = array(
        'Search.Searchable',
        'Upload.Upload' => array(
            'company_logo' => array(
                'fields' => array(
                    'dir' => 'company_logo_dir'
                ),
                'thumbnailSizes' => array(
                    'thumb' => '130x130'
                ),
                'thumbnailMethod' => 'php',
                'deleteOnUpdate' => true,
            )
        )
    );
    public $filterArgs = array(
        'name' => array(
            'type' => 'like',
            'field' => 'name'
        ),
        'username' => array(
            'type' => 'like',
            'field' => 'username'
        ),
        'email' => array(
            'type' => 'like',
            'field' => 'email'
        ),
        'company_name' => array(
            'type' => 'like',
            'field' => 'company_name'
        ),
        'term' => array(
            'type' => 'query',
            'method' => 'searchTermCondition'
        )
    );

    public function searchTermCondition($data = array())
    {
        $filter = $data['term'];
        $conditions = array(
            'OR' => array(
                $this->alias . '.name LIKE' => '%' . $this->formatLike($filter) . '%',
                $this->alias . '.username LIKE' => '%' . $this->formatLike($filter) . '%',
                $this->alias . '.email LIKE' => '%' . $this->formatLike($filter) . '%',
            )
        );
        return $conditions;
    }

    /**
     * Validation rules
     *
     * @var array
     */
    public $validate = array(
        'password' => array(
            'notEmpty' => array(
                'rule' => array('notEmpty'),
                'message' => 'Password is required',
                'on' => 'create',
            ),
        ),
        'password2' => array(
            'notEmpty' => array(
                'rule' => array('notEmpty'),
                'message' => 'Password confirmation is required',
                'allowEmpty' => true
            ),
            'compare' => array(
                'rule' => array('validate_passwords'),
                'message' => 'The passwords you entered do not match.',
                'allowEmpty' => true
            )
        ),
        'name' => array(
            'notEmpty' => array(
                'rule' => array('notEmpty'),
                'message' => 'Name is required',
                'allowEmpty' => true
            ),
        ),
        'email' => array(
            'notEmpty' => array(
                'rule' => array('notEmpty'),
                'message' => 'Email is required',
                'on' => 'create', // Limit validation to 'create' or 'update' operations
            ),
            'email' => array(
                'rule' => array('email'),
                'message' => 'Please enter valid email',
                'on' => 'create', // Limit validation to 'create' or 'update' operations
            ),
            'isUnique' => array(
                'rule' => array('isUnique'),
                'message' => 'This email has already been taken',
                'on' => 'create',
            ),
        ),
        'company_logo' => array(
            'isUnderPhpSizeLimit' => array(
                'rule' => 'isUnderPhpSizeLimit',
                'message' => 'File exceeds upload filesize limit'
            ),
            'isUnderFormSizeLimit' => array(
                'rule' => 'isUnderFormSizeLimit',
                'message' => 'File exceeds form upload filesize limit'
            ),
            'isCompletedUpload' => array(
                'rule' => 'isCompletedUpload',
                'message' => 'File was not successfully uploaded'
            ),
            'isValidExtension' => array(
                'rule' => array('isValidExtension', array('jpeg', 'png', 'gif', 'jpg'), false),
                'message' => 'File does not have a pdf, png, or txt extension'
            ),
            'isBelowMaxSize' => array(
                'rule' => array('isBelowMaxSize', 5242880, false),
                'message' => 'File is larger than the maximum filesize (5mb)'
            ),
            'isAboveMinSize' => array(
                'rule' => array('isAboveMinSize', 1024, false),
                'message' => 'File is below the mimimum filesize'
            )
        )
    );
    
    public $hasMany = array(
        'Job' => array(
            'className' => 'Job',
            'foreignKey' => 'user_id',
            'conditions' => '',
            'order' => 'Job.created DESC',
            'limit' => '',
            'dependent' => true
        )
    );

    public function validate_passwords()
    {
        return $this->data[$this->alias]['password'] === $this->data[$this->alias]['password2'];
    }

    public function beforeSave($options = array())
    {
        if (isset($this->data[$this->alias]['password'])) {
            $passwordHasher = new BlowfishPasswordHasher();
            $this->data[$this->alias]['password'] = $passwordHasher->hash(
                $this->data[$this->alias]['password']
            );
        }

        if (!empty($this->data['User']['expiry_date'])) {
            $this->data['User']['expiry_date'] = $this->dateTimeFormatBeforeSave(
                $this->data['User']['expiry_date']
            );
        }

        return true;
    }

    public function generateHashChangePassword()
    {
        $salt = Configure::read('Security.salt');
        $salt_v2 = Configure::read('Security.cipherSeed');
        $rand = mt_rand(1, 999999999);
        $rand_v2 = mt_rand(1, 999999999);

        $hash = hash('sha256', $salt . $rand . $salt_v2 . $rand_v2);

        return $hash;
    }

    public function afterFind($results, $primary = false)
    {
        foreach ($results as $key => $val) {
            if (isset($val['User']['expiry_date'])) {
                $days_to_expired = $this->calculateDaysToExpired($val['User']['expiry_date']);

                if ($days_to_expired > 0) {
                    $results[$key]['User']['is_expired'] = false;
                } else {
                    $results[$key]['User']['is_expired'] = true;
                }

                $results[$key]['User']['days_to_expired'] = $days_to_expired;

                $results[$key]['User']['expiry_date'] = date("d-M-Y g:i A", strtotime($results[$key]['User']['expiry_date']));
            } else {
                $results[$key]['User']['is_expired'] = true;
            }
        }
        return $results;
    }

    public function calculateDaysToExpired($expiry_date)
    {
        $now = time();
        $expiryDate = strtotime($expiry_date);
        $datediff = $expiryDate - $now;
        $days_old = floor($datediff / (60 * 60 * 24));
        return $days_old + 1;
    }
}
